from openTSNE import TSNE
import pandas as pd
import pickle
import _pickle as cPickle
import bz2
from sklearn.feature_selection import VarianceThreshold
import hdbscan

def compressed_pickle(title, data):
    with bz2.BZ2File(title + '.pbz2', 'w') as f:
        cPickle.dump(data, f)


def tsne_train(prefix, outfolder, min_cluster, threads, perplex, exagger, variance):
    sorted_filename = "sorted_" + prefix + ".rtab"
    filepath = (outfolder + '/' + sorted_filename)
    print("Loading unitigs...")
    gene_presence_absence = pd.read_table(filepath)
    gene_presence_absence = gene_presence_absence.reset_index(drop=True)
    print("Unitigs loaded")

    new_df = gene_presence_absence.drop(['Unitig_sequence'], axis = 1)

    transpose_df = new_df.transpose()


    vt =  VarianceThreshold(threshold=(variance))
    _ = vt.fit(transpose_df)
    mask = vt.get_support()

    transpose_df = transpose_df.loc[:, mask]

    transpose_df['sum'] = transpose_df.sum(axis=1)

    transpose_df = transpose_df.loc[transpose_df['sum'] > 0]

    transpose_df = transpose_df.drop(['sum'], axis = 1)

    selected_unitigs = gene_presence_absence['Unitig_sequence'][mask]
    #selected_unitigs = gene_presence_absence['Unitig_sequence']
    path_selected_unitigs = (outfolder + '/' + prefix + '_unitigs.csv')
    selected_unitigs.to_csv(path_selected_unitigs)

    numpy_presence_absence = transpose_df.to_numpy()

    tsne = TSNE(
        perplexity=perplex,
        exaggeration=exagger,
        metric="jaccard",
        n_jobs=threads,
        random_state=42,
        neighbors='exact',
        verbose=True,
    )

    print("Training opentsne")

    embedding_train = tsne.fit(numpy_presence_absence)

    embedding_query = embedding_train.transform(numpy_presence_absence)

    df_tsne_query = pd.DataFrame(embedding_query)

    path_model = (outfolder + '/' + prefix + '_embedding')

    compressed_pickle(path_model, embedding_train)

    print("Saving results")

    df_tsne = pd.DataFrame(embedding_train)

    df_tsne_query = pd.DataFrame(embedding_query)

    path_query = (outfolder + '/' + prefix + '_sanity_check.csv')

    df_tsne_query.to_csv(path_query)

    hdbscan_clusterer = hdbscan.HDBSCAN(min_cluster_size = min_cluster, min_samples = min_cluster, prediction_data=True)

    hdbscan_clusterer.fit(df_tsne)

    path_hdbscan_model = (outfolder + '/' + prefix + '_clusters')

    compressed_pickle(path_hdbscan_model, hdbscan_clusterer)

    hdbscan_information = pd.DataFrame(hdbscan_clusterer.labels_)

    hdbscan_probabilities = pd.DataFrame(hdbscan_clusterer.probabilities_)

    opentsne_info = pd.concat([df_tsne, hdbscan_information, hdbscan_probabilities, pd.DataFrame(transpose_df.index.values)], axis=1)
    opentsne_info = opentsne_info.reset_index(drop=True)
    opentsne_info.columns = ['tsne1D', 'tsne2D', 'Standard_Cluster', 'Membership_Probability','Sample_Name']

    path_results_model = (outfolder + '/' + prefix + '_results.csv')

    opentsne_info.to_csv(path_results_model, index = False)
